#pragma TextEncoding	= "UTF-8"
#pragma rtGlobals		= 3
#pragma IgorVersion		= 7.00
#pragma moduleName		= AnalysisTools
#pragma version			= 1.07
#pragma DefaultTab		= {3,20,4}

static Constant kVersion = 1.07
static StrConstant kVersionDate = "11/2022"

// --------------------- Project Updater header ----------------------
// If you're using Igor Pro 8 or later and have Project Updater
// installed, this package can check periodically for new releases.
// https://www.wavemetrics.com/project/Updater
static Constant kProjectID = 21866																// the project node on IgorExchange
static StrConstant ksShortTitle = "Super Quick Fit"												// the project short title on IgorExchange

//________________________________________________________________________________________________
//										Super Quick Fit
//	Written by Stephan Thuermer - https://www.wavemetrics.com/user/chozo
//
//	Provides a new right-click menu similar to the built-in Quick Fit. Several custom options have
//	been added for quick access and user-defined functions are supported. To add your own function
//	you need to provide:
//		- a standard or all-at-once fit function with an unique name
//		- a guess function with the name fitfuncName+CoefGen_SUFFIX
//	
//	Optionally, derived results can be printed from another function named fitfuncName+derivedVal_SUFFIX
//
//	Both the guess and derived-results functions have the SuperQuickFitStruct as input.
//	The guess function is expected to generate a coefficient wave with the correct number of
//	parameters and return this wave to the s.cw parameter. Additionally, the s.holdStr can be set
//	to hold one parameter (usually the baseline) if the option is activated in the menu.
// 
//	Additional custom fit functions are provided with the package:
//		- Fermi-Dirac function with linear slope
//		- Line with x-crossing as input parameter (instead of y-crossing)
//		- Angular Distribution function (e.g., for Photoelectron Angular Distributions)
//		- Double Gauss function
//		- Exponentially Modified Gaussian
//
//	Support for standard fit functions with the extended options is provided as well.
//
// 2022-04-07 - ver. 1.00:	Initial public release.
// 2022-04-08 - ver. 1.01:	Added textbox preferences.
// 2022-05-11 - ver. 1.02:	Fixed bug: Missing M_Covar for Line fit was not handled properly.
// 2022-07-14 - ver. 1.03:	Improved guess for EF parameter in Fermi-Line guess function.
//							Improved derived parameters for EMG peak shape.
// 2022-10-18 - ver. 1.04:	'Clean-up' now only removes fit-related variables.
//							Derived values will be added to wave notes as well.
//							Improved guess for DoubleGauss function.
// 2022-10-27 - ver. 1.05:	Added ExpConvExp function (thanks to Oleg Dyachok from Uppsala University for the idea).
//							Now negative EMG peaks can be fitted.
//							Fixed bug: 'use cursor' was not working for small delta-scaling.
// 2022-11-11 - ver. 1.06:	Fixed bug: EMG and ExpConvExp guesses did not work for XY data when the y-wave was scaled.
//							Added peak-area output to EMG peak.
//							Added option to align the fit trace to the source trace (thanks to Oleg Dyachok from Uppsala University for the idea).
// 2022-11-27 - ver. 1.07:	Added support for fitting traces from 2D waves.
//							Added additional derived parameters for ExpConvExp (thanks to Oleg Dyachok for the help).
//							Added confidence intervals to the area of the ExpModGauss and ExpConvExp functions.
//							Changed behavior: Now the cursor placement does not set the baseline, but only decides the fit range.
//________________________________________________________________________________________________

strConstant QuickFit_DefaultFuncs1D = "Gauss;Lor;Voigt;Sin;Line;-;HillEquation;Sigmoid;Power;LogNormal;Log;-;Exp;Exp_xOffset;DblExp;DblExp_Peak;DblExp_xOffset;"
strConstant QuickFit_DefaultFuncs2D = "Gauss2D;"

static strConstant CoefGen_SUFFIX = "_prepareCoef"												// name suffixes for coef and output functions 
static strConstant derivedVal_SUFFIX = "_derivedVals"

Function dummyFunc_prepareCoef(STRUCT SuperQuickFitStruct &s)									// dummy functions for coef and output
	return -1
End
Function/S dummyFunc_derivedVals(STRUCT SuperQuickFitStruct &s)
	return ""
End

Structure SuperQuickFitStruct
	Wave cw			// coef wave
	Wave sw			// covariance matrix (M_Covar)
	Wave data		// data wave (1D or 2D)
	Wave xw			// x wave
	Wave yw			// y wave (for images; not supported yet)
	int32 pntMin	// start point of fit range
	int32 pntMax	// end point of fit range
	int32 pntMinY	// start point in y direction (for images)
	int32 pntMaxY	// end point  in y direction (for images)
	int32 pntCol	// column to use for traces from 2D wave
	double xBase	// x baseline / start offset
	double yBase	// y baseline / start offset
	int16 holdMode	// current hold setting: [0] = no hold, [1] = force o zero, [2] = force to min
	int16 plotFull	// setting to plot over full range
	int16 doTextbox	// setting for text-box drawing
	String holdStr	// hold string for the fit
EndStructure

// +++++++++++++++++++++++++++++ persistent preferences support ++++++++++++++++++++++++++++++++++

static strConstant defaultSettings	= "offsetMode:0;plotFull:0;fitRangeMode:1;addTextBox:0;weightError:1;outCleanup:0;textboxOpt:791;plotAligned:0;"
static strConstant kPackageName		= "SuperQuickFit"
static strConstant kPrefsFileName	= "SuperQuickFitPrefs.bin"
static Constant kPrefsVersion		= 100

Structure SuperQuickFitPrefs
	uint32	version					// Preferences structure version number. 100 means 1.00.
	int16	offsetMode
	int16	plotFull
	int16	fitRangeMode
	int16	addTextBox
	int16	weightError
	int16	outCleanup
	int16	textboxOpt
	int16	plotAligned				// new since version 1.06
	char reserved[200 - 8*2 - 4]	// Reserved for future use
EndStructure

static Function/S LoadPrefs(STRUCT SuperQuickFitPrefs &s)
	LoadPackagePreferences kPackageName, kPrefsFileName, 0, s
	String settings = ""
	if (V_flag!=0 || V_bytesRead==0 || s.version!=kPrefsVersion)								// If error or prefs not found or not valid, initialize them.
		AnalysisTools#SavePrefs(s, defaultSettings)
	endif
	sPrintf settings, "offsetMode:%d;plotFull:%d;fitRangeMode:%d;addTextBox:%d;weightError:%d;outCleanup:%d;textboxOpt:%d;plotAligned:%d;", s.offsetMode, s.plotFull, s.fitRangeMode, s.addTextBox, s.weightError, s.outCleanup, s.textboxOpt, s.plotAligned
	return settings
End

static Function SavePrefs(STRUCT SuperQuickFitPrefs &s, String settings)
	AnalysisTools#WritePrefs(s, settings)
	SavePackagePreferences kPackageName, kPrefsFileName, 0, s
End

static Function WritePrefs(STRUCT SuperQuickFitPrefs &s, String settings)
	s.version		= kPrefsVersion
	s.offsetMode	= NumberByKey("offsetMode" ,settings)
	s.plotFull		= NumberByKey("plotFull" ,settings)
	s.fitRangeMode	= NumberByKey("fitRangeMode" ,settings)
	s.addTextBox	= NumberByKey("addTextBox" ,settings)
	s.weightError	= NumberByKey("weightError" ,settings)
	s.outCleanup	= NumberByKey("outCleanup" ,settings)
	s.textboxOpt	= NumberByKey("textboxOpt" ,settings)
	s.plotAligned	= NumberByKey("plotAligned" ,settings)
End

//################################################################################################

Menu "GraphPopup", dynamic
	SubMenu "Super Quick Fit"
		SubMenu "Settings"
			AnalysisTools#QuickFitMenuItem("useCursor"),	/Q, AnalysisTools#QuickFitMenuSettings("useCursor")
			AnalysisTools#QuickFitMenuItem("useAxRange"),	/Q, AnalysisTools#QuickFitMenuSettings("useAxRange")
			AnalysisTools#QuickFitMenuItem("zeroBase"),		/Q, AnalysisTools#QuickFitMenuSettings("zeroBase")
			AnalysisTools#QuickFitMenuItem("fixedBase"),	/Q, AnalysisTools#QuickFitMenuSettings("fixedBase")
			AnalysisTools#QuickFitMenuItem("weightError"),	/Q, AnalysisTools#QuickFitMenuSettings("weightError")
			AnalysisTools#QuickFitMenuItem("outCleanup"),	/Q, AnalysisTools#QuickFitMenuSettings("outCleanup")
			"-"
			AnalysisTools#QuickFitMenuItem("plotFull"),		/Q, AnalysisTools#QuickFitMenuSettings("plotFull")
			AnalysisTools#QuickFitMenuItem("plotAligned"),	/Q, AnalysisTools#QuickFitMenuSettings("plotAligned")
			AnalysisTools#QuickFitMenuItem("addTextBox"),	/Q, AnalysisTools#QuickFitMenuSettings("addTextBox")
			"\t»Textbox Preferences ...",					/Q, AnalysisTools#QickFitTextBoxOptions()
		End
		"-"
		AnalysisTools#QuickFitMenuUserEntries(), /Q, AnalysisTools#SuperQuickFit("user")
		"-"
		SubMenu AnalysisTools#QuickFitMenuSubMenuEntries("Poly")
			"1;2;3;4;5;6;7;8;9;10;", /Q, AnalysisTools#SuperQuickFit("poly")
		End
		SubMenu AnalysisTools#QuickFitMenuSubMenuEntries("Poly_xOffset")
			"1;2;3;4;5;6;7;8;9;10;", /Q, AnalysisTools#SuperQuickFit("poly_XOffset")
		End
		SubMenu AnalysisTools#QuickFitMenuSubMenuEntries("Poly2D")
			"1;2;3;4;5;6;7;8;9;10;", /Q, AnalysisTools#SuperQuickFit("Poly2D")
		End
		"-"
		AnalysisTools#QuickFitMenuDefaultEntries(), /Q, AnalysisTools#SuperQuickFit("")
	End
End
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Menu "TracePopup", dynamic
	SubMenu "Super Quick Fit"
		SubMenu "Settings"
			AnalysisTools#QuickFitMenuItem("useCursor"),	/Q, AnalysisTools#QuickFitMenuSettings("useCursor")
			AnalysisTools#QuickFitMenuItem("useAxRange"),	/Q, AnalysisTools#QuickFitMenuSettings("useAxRange")
			AnalysisTools#QuickFitMenuItem("zeroBase"),		/Q, AnalysisTools#QuickFitMenuSettings("zeroBase")
			AnalysisTools#QuickFitMenuItem("fixedBase"),	/Q, AnalysisTools#QuickFitMenuSettings("fixedBase")
			AnalysisTools#QuickFitMenuItem("weightError"),	/Q, AnalysisTools#QuickFitMenuSettings("weightError")
			AnalysisTools#QuickFitMenuItem("outCleanup"),	/Q, AnalysisTools#QuickFitMenuSettings("outCleanup")
			"-"
			AnalysisTools#QuickFitMenuItem("plotFull"),		/Q, AnalysisTools#QuickFitMenuSettings("plotFull")
			AnalysisTools#QuickFitMenuItem("plotAligned"),	/Q, AnalysisTools#QuickFitMenuSettings("plotAligned")
			AnalysisTools#QuickFitMenuItem("addTextBox"),	/Q, AnalysisTools#QuickFitMenuSettings("addTextBox")
			"\t> Textbox Preferences ...",					/Q, AnalysisTools#QickFitTextBoxOptions()
		End
		"-"
		AnalysisTools#QuickFitMenuUserEntries(), /Q, AnalysisTools#SuperQuickFit("user")
		"-"
		SubMenu AnalysisTools#QuickFitMenuSubMenuEntries("Poly")
			"1;2;3;4;5;6;7;8;9;10;", /Q, AnalysisTools#SuperQuickFit("poly")
		End
		SubMenu AnalysisTools#QuickFitMenuSubMenuEntries("Poly_xOffset")
			"1;2;3;4;5;6;7;8;9;10;", /Q, AnalysisTools#SuperQuickFit("poly_XOffset")
		End
		SubMenu AnalysisTools#QuickFitMenuSubMenuEntries("Poly2D")
			"1;2;3;4;5;6;7;8;9;10;", /Q, AnalysisTools#SuperQuickFit("Poly2D")
		End
		"-"
		AnalysisTools#QuickFitMenuDefaultEntries(), /Q, AnalysisTools#SuperQuickFit("")
	End
End

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static Function/S QuickFitMenuSubMenuEntries(String entry)										// sub menu entry generator
	int noTraces = 0, notCompatible = 0
	AnalysisTools#checkTraceOrImage(noTraces, notCompatible)
	if (noTraces)
		return SelectString(StringMatch(entry,"*2D") && !notCompatible,"(","")+entry
	else
		return SelectString(StringMatch(entry,"*2D") && !notCompatible,"","(")+entry
	endif
End

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static Function/S QuickFitMenuUserEntries()														// entries from user functions
	String funcList = RemoveFromList("dummyFunc",ReplaceString(CoefGen_SUFFIX,FunctionList("*"+CoefGen_SUFFIX, ";", ""),""))
	int noTraces = 0, notCompatible = 0
	AnalysisTools#checkTraceOrImage(noTraces, notCompatible)
	return AnalysisTools#deactivateList(funcList, noTraces)
End

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static Function/S QuickFitMenuDefaultEntries()													// default fit list
	int noTraces = 0, notCompatible = 0
	AnalysisTools#checkTraceOrImage(noTraces, notCompatible)
	return AnalysisTools#deactivateList(QuickFit_DefaultFuncs2D, !noTraces || notCompatible) + AnalysisTools#deactivateList(QuickFit_DefaultFuncs1D, noTraces)
End

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static Function/S getTopGraph()
	return StringFromList(0,WinList("*", ";", "WIN:1"))
End

static Function checkTraceOrImage(int &noTraces, int &notCompatible)							// checks for compatible image or trace in top graph
	String gName = getTopGraph()
	if (!strlen(gName))
		return -1
	endif
	String imgInfo = ImageInfo(gName,"",0)
	noTraces = ItemsInList(TraceNameList(gName,";",1)) == 0
	if (strlen(imgInfo) && noTraces)
		Wave/Z imgXW = $(StringByKey("XWAVEDF",imgInfo)+StringByKey("XWAVE",imgInfo))
		Wave/Z imgYW = $(StringByKey("YWAVEDF",imgInfo)+StringByKey("YWAVE",imgInfo))									
		notCompatible = WaveExists(imgXW) || WaveExists(imgYW)									// no support for x/y waves with images yet
	endif
	return 0
End

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static Function/S deactivateList(String list, int deactivate)									// deactivate menu entry list (prepend '(' character)
	if (deactivate)
		return "("+RemoveEnding(ReplaceString(";", list, ";("),"(")
	else
		return list
	endif
End

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static Function/S QuickFitMenuItem(String which)												// returns menu entry for settings
	STRUCT SuperQuickFitPrefs prefs
	String settings = StrVarOrDefault("root:QuickFitMenu_settings",AnalysisTools#LoadPrefs(prefs))
	StrSwitch(which)
		case "zeroBase":
			return SelectString(NumberByKey("offsetMode"	,settings) == 1,"","!"+num2char(18)) + "Offset: Force to Zero"
		case "fixedBase":
			return SelectString(NumberByKey("offsetMode"	,settings) == 2,"","!"+num2char(18)) + "Offset: Force to Minimum"
		case "plotFull":
			return SelectString(NumberByKey("plotFull"		,settings) == 1,"","!"+num2char(18)) + "Result: Plot Full Range"
		case "plotAligned":
			return SelectString(NumberByKey("plotAligned"	,settings) == 1,"","!"+num2char(18)) + "Result: Align Fit with Trace"
		case "addTextBox":
			return SelectString(NumberByKey("addTextBox"	,settings) == 1,"","!"+num2char(18)) + "Result: Add TextBox"
		case "useCursor":
			return SelectString(NumberByKey("fitRangeMode"	,settings) == 1,"","!"+num2char(18)) + "Fit: Between Cursors"
		case "useAxRange":
			return SelectString(NumberByKey("fitRangeMode"	,settings) == 2,"","!"+num2char(18)) + "Fit: Within Axis Range"
		case "weightError":
			return SelectString(NumberByKey("weightError"	,settings) == 1,"","!"+num2char(18)) + "Weight from Error Bar Wave"
		case "outCleanup":
			return SelectString(NumberByKey("outCleanup"	,settings) == 1,"","!"+num2char(18)) + "Clean up Output Waves + Variables"
	EndSwitch
End

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static Function/S QuickFitMenuSettings(String which)											// write settings from selection
	STRUCT SuperQuickFitPrefs prefs
	String settings = StrVarOrDefault("root:QuickFitMenu_settings",AnalysisTools#LoadPrefs(prefs))
	StrSwitch(which)
		case "zeroBase":
			settings = ReplaceNumberByKey("offsetMode"	,settings ,(NumberByKey("offsetMode"	,settings) == 1 ? 0 : 1))
		break
		case "fixedBase":
			settings = ReplaceNumberByKey("offsetMode"	,settings ,(NumberByKey("offsetMode"	,settings) == 2 ? 0 : 2))
		break
		case "useCursor":
			settings = ReplaceNumberByKey("fitRangeMode",settings ,(NumberByKey("fitRangeMode"	,settings) == 1 ? 0 : 1))
		break
		case "useAxRange":
			settings = ReplaceNumberByKey("fitRangeMode",settings ,(NumberByKey("fitRangeMode"	,settings) == 2 ? 0 : 2))
		break
		default:
			settings = ReplaceNumberByKey(which	,settings ,(NumberByKey(which ,settings) == 1 ? 0 : 1))
		break
	EndSwitch
	String/G root:QuickFitMenu_settings = settings
	AnalysisTools#SavePrefs(prefs, settings)
End

//##################################### Text Options Panel ######################################

static StrConstant kOptBits = "Title;Date;Time;FitType;FitFunc;ModelWave;YWave;XWave;CoefValues;CoefErrors;"

static Function QickFitTextBoxOptions()
	if (WinType("SQF_TextBoxOpt"))
		return 0
	endif
	String optTitle = "Select All;Title;Date and Time;Date;Time;Fit Type;Fit Function Name;Waves;Model Wave;Y Wave;X Wave(s);Coefficient Report;Coefficient Values;Coefficient Errors;"
	String optNames = "All;Title;DateTime;Date;Time;FitType;FitFunc;Waves;ModelWave;YWave;XWave;CoefReport;CoefValues;CoefErrors;"
	String optAlign = "0;0;0;15;15;0;0;0;15;15;15;0;15;15;"
	GetMouse
	NewPanel/FLT=2/K=1/W=(V_left,V_top,V_left+155,V_top+400)/N=SQF_TextBoxOpt as "Textbox Prefs"
	int i, dist = 25
	for (i=0; i<ItemsInList(optTitle); i++)
		CheckBox $("c"+StringFromList(i,optNames)) ,pos={15+str2num(StringFromList(i,optAlign)),15+dist*i} ,size={100,23} ,title=StringFromList(i,optTitle) ,proc=AnalysisTools#SQF_TextboxOptions
	endfor
	Button bSave	,pos={10 ,15+dist*i} ,size={60,22} ,title="Save"	,proc=AnalysisTools#SQF_TextBoxButton
	Button bCancel	,pos={85 ,15+dist*i} ,size={60,22} ,title="Cancel"	,proc=AnalysisTools#SQF_TextBoxButton
	AnalysisTools#SQF_TextBoxLoadFromSettings()
	return 0
End

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static Function SQF_TextboxOptions(s) : CheckBoxControl
	STRUCT WMCheckboxAction &s
	if(s.eventCode == 2)
		String ctrlList = ""
		StrSwitch(s.ctrlName)
			case "cDateTime":
				ctrlList = "cDate;cTime;"
			break
			case "cWaves":
				ctrlList = ControlNameList(s.win, ";", "*Wave")
			break
			case "cCoefReport":
				ctrlList = ControlNameList(s.win, ";", "cCoef*")
			break
			case "cAll":
				ctrlList = ControlNameList(s.win, ";", "c*")
			break
		EndSwitch
		int i, count = 0
		for (i=0; i<ItemsInList(ctrlList); i++)
			CheckBox $(StringFromList(i,ctrlList)) ,win=SQF_TextBoxOpt ,value=s.checked
		endfor
		ctrlList = RemoveFromList("cAll", ControlNameList(s.win, ";", "c*"))
		for (i=0; i<ItemsInList(ctrlList); i++)
			ControlInfo/W=SQF_TextBoxOpt $("c"+StringFromList(i,kOptBits)); count += V_value
		endfor
		CheckBox cAll ,win=SQF_TextBoxOpt ,value=(ItemsInList(ctrlList) == count)
	endif
	return 0
End

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static Function SQF_TextBoxButton(s) : ButtonControl
	STRUCT WMButtonAction &s
	if (s.eventCode == 2)
		if(CmpStr(s.ctrlName,"bSave") == 0)
			AnalysisTools#SQF_TextBoxWriteSettings()											// save options
		endif
		KillWindow SQF_TextBoxOpt
	endif
	return 0
End

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static Function SQF_TextBoxWriteSettings()
	STRUCT SuperQuickFitPrefs prefs
	String settings = StrVarOrDefault("root:QuickFitMenu_settings",AnalysisTools#LoadPrefs(prefs))
	Variable i, textboxOpt = 0
	for (i=0; i<ItemsInList(kOptBits); i++)														// read checked bits from checkboxes
		ControlInfo/W=SQF_TextBoxOpt $("c"+StringFromList(i,kOptBits))
		textboxOpt = V_Value ? textboxOpt | (2^i) : textboxOpt
	endfor
	settings = ReplaceNumberByKey("textboxOpt",settings,textboxOpt)								// write new option
	String/G root:QuickFitMenu_settings = settings												// save settings
	AnalysisTools#SavePrefs(prefs, settings)
	return 0
End

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static Function SQF_TextBoxLoadFromSettings()													// transfers checked bits to checkboxes
	STRUCT SuperQuickFitPrefs prefs
	String settings = StrVarOrDefault("root:QuickFitMenu_settings",AnalysisTools#LoadPrefs(prefs))
	Variable i, textboxOpt = NumberByKey("textboxOpt",settings)
	for (i=0; i<ItemsInList(kOptBits); i++)
		CheckBox $("c"+StringFromList(i,kOptBits)) ,win=SQF_TextBoxOpt ,value=((textboxOpt & 2^i) != 0)
	endfor
	CheckBox cDateTime		,win=SQF_TextBoxOpt ,value=((textboxOpt & 2^1) != 0 && (textboxOpt & 2^2) != 0)
	CheckBox cWaves			,win=SQF_TextBoxOpt ,value=((textboxOpt & 2^5) != 0 && (textboxOpt & 2^6) != 0 && (textboxOpt & 2^7) != 0)
	CheckBox cCoefReport	,win=SQF_TextBoxOpt ,value=((textboxOpt & 2^8) != 0 && (textboxOpt & 2^9) != 0)
	CheckBox cAll			,win=SQF_TextBoxOpt ,value=( textboxOpt == 1023)
	return 0
End

//________________________________________________________________________________________________
//									Super Quick Fit Kernel
//________________________________________________________________________________________________

static Function SuperQuickFit(String option)
	STRUCT SuperQuickFitStruct s
	STRUCT SuperQuickFitPrefs prefs
	
	GetLastUserMenuInfo																			// figure out graph and trace names
	String gName = S_graphName, trace = S_traceName, fitFunc = S_value, fitType = "CurveFit"	// grab fitFunc from selected item
	if (StringMatch(option, "poly*"))															// for poly fits
		fitFunc = option + " " + S_value + ","
	endif
	if (StringMatch(option, "user"))
		fitType = "FuncFit"
		if (!Exists(fitFunc))
			return -1
		endif
	endif
	if (!strlen(fitFunc))
		return -1
	endif
	
	String imgInfo = ImageInfo(gName,"",0)
	Variable isImage = strlen(imgInfo), col2D = NaN
	if (!strlen(trace))																			// was called from graph popup => find top trace
		trace = StringFromList(0, TraceNameList(gName,";",1))
	endif
	if ((!strlen(trace) && !isImage)|| !strlen(fitFunc))										// still no trace or function => abort
		return -1
	endif
	
	if (isImage)
		Wave s.data = $(StringByKey("ZWAVEDF",imgInfo)+StringByKey("ZWAVE",imgInfo))
		Wave/Z s.xw = $(StringByKey("XWAVEDF",imgInfo)+StringByKey("XWAVE",imgInfo))
		Wave/Z s.yw = $(StringByKey("YWAVEDF",imgInfo)+StringByKey("YWAVE",imgInfo))
		if (WaveExists(s.xw) || WaveExists(s.yw))												// no support for x/y waves with images yet
			return -1
		endif
		s.pntMinY = 0
		s.pntMaxY = DimSize(s.data,1)-1
	else
		Wave s.data = TraceNameToWaveRef(gName, trace)
		Wave/Z s.xw = XWaveRefFromTrace(gName, trace)
		if (WaveDims(s.data) == 2)																// for traces of 2D data
			String trRange = StringByKey("YRANGE", TraceInfo(gName,trace,0))
			trRange = trRange[strsearch(trRange,"[",inf,1)+1,inf]								// find the displayed column from TraceInfo
			col2D = str2num(RemoveEnding(trRange,"]"))											// will be appended to the fit command
		endif
	endif
	int isTraceOf2D = numtype(col2D) == 0
	s.pntCol = isTraceOf2D ? col2D : 0
	s.pntMin = 0
	s.pntMax = DimSize(s.data,0)-1
	
	String settings = StrVarOrDefault("root:QuickFitMenu_settings",AnalysisTools#LoadPrefs(prefs))	// load settings
	String/G root:QuickFitMenu_settings = settings												// save used settings to experiment
	int textboxOpt	= NumberByKey("textboxOpt",settings)
	int rangeMode	= NumberByKey("fitRangeMode",settings)
	int doWeight	= NumberByKey("weightError",settings)
	int doCleanUp	= NumberByKey("outCleanup",settings)
	int doFitAlign	= NumberByKey("plotAligned",settings)
	doFitAlign = doFitAlign < 0 ? 0 : doFitAlign												// this option was introduced in ver. 1.06 and may not be initialized yet
	s.plotFull	= NumberByKey("plotFull",settings)
	s.doTextbox = NumberByKey("addTextBox",settings)
	s.holdMode  = NumberByKey("offsetMode",settings)
	s.holdStr 	= "1"																			// default is baseline hold
	// +++++++++++++++++++++++++ find fit range and baseline values ++++++++++++++++++++++++++++++	
	int i = 0
	String csrA = "", csrB = "", csrList = "A;B;C;D;E;F;G;H;I;J;"
	Variable csrAx = NaN, csrBx = NaN, csrAy = NaN, csrBy = NaN
	for (; i < ItemsInList(csrList);)															// see which cursor is placed
		csrA = StringFromList(i++,csrList)
		if (strlen(CsrInfo($csrA,gName)))
			csrAx = hcsr($csrA, gName)
			csrAy = vcsr($csrA, gName)
			break
		endif
	endfor
	for (; i < ItemsInList(csrList);)															// poll the second cursor
		csrB = StringFromList(i++,csrList)
		if (strlen(CsrInfo($csrB,gName)))
			csrBx = hcsr($csrB, gName)
			csrBy = vcsr($csrB, gName)
			break
		endif
	endfor
	
	Variable xMin = NaN, xMax = NaN, yMin = NaN, yMax = NaN, temp
	Variable useCsr = numtype(csrAx) == 0 && numtype(csrBx) == 0 && csrAx != csrBx && rangeMode == 1
	if (rangeMode > 0)
		if (useCsr)
			xMin = csrAx; xMax = csrBx; yMin = csrAy; yMax = csrBy
		elseif (rangeMode == 2)
				if (WhichListItem("bottom",AxisList(gName)) > -1)
					GetAxis/W=$gName/Q bottom
				elseif (WhichListItem("top",AxisList(gName)) > -1)
					GetAxis/W=$gName/Q top
				endif
				xMin = !V_flag ? V_min : xMin
				xMax = !V_flag ? V_max : xMax
			if (isImage)
				if (WhichListItem("left",AxisList(gName)) > -1)
					GetAxis/W=$gName/Q left
				elseif (WhichListItem("right",AxisList(gName)) > -1)
					GetAxis/W=$gName/Q right
				endif
				yMin = !V_flag ? V_min : yMin
				yMax = !V_flag ? V_max : yMax
			endif
		endif
		if (numtype(xMin) == 0 && numtype(xMax) == 0)
			if (WaveExists(s.xw))
				FindLevel/P/Q s.xw, xMin; s.pntMin = (!v_flag && v_levelX > 0 && v_levelX < DimSize(s.data,0)-1) ? v_levelX : s.pntMin
				FindLevel/P/Q s.xw, xMax; s.pntMax = (!v_flag && v_levelX > 0 && v_levelX < DimSize(s.data,0)-1) ? v_levelX : s.pntMax
			else
				s.pntMin = limit(x2pnt(s.data,xMin), 0, DimSize(s.data,0)-1)
				s.pntMax = limit(x2pnt(s.data,xMax), 0, DimSize(s.data,0)-1)
			endif
		endif
		if (isImage && numtype(yMin) == 0 && numtype(yMax) == 0)
			s.pntMinY = limit(ScaleToIndex(s.data,yMin,1), 0, DimSize(s.data,1)-1)
			s.pntMaxY = limit(ScaleToIndex(s.data,yMax,1), 0, DimSize(s.data,1)-1)
		endif
	endif
	
	if (s.pntMin > s.pntMax)
		temp = s.pntMin
		s.pntMin = s.pntMax
		s.pntMax = temp
	endif
	if (s.pntMinY > s.pntMaxY)
		temp = s.pntMinY
		s.pntMinY = s.pntMaxY
		s.pntMaxY = temp
	endif
	
	Variable reverseX = 1
	if (WaveExists(s.xw))
		s.xBase = s.xw[0]
		reverseX = sign(s.xw[numpnts(s.xw)-1]-s.xw[0])
	else
		s.xBase  = DimOffset(s.data,0)
		reverseX = sign(DimDelta(s.data,0))
	endif
	
	if (isImage)
		WaveStats/Q/RMD=[s.pntMin,s.pntMax][s.pntMinY,s.pntMaxY] s.data
	else
		WaveStats/Q/RMD=[s.pntMin,s.pntMax][s.pntCol] s.data
	endif
	s.yBase = V_max > V_min ? V_min : V_max														// the baseline value
	if (useCsr)
			s.xBase = hcsr($csrA, gName)*reverseX < hcsr($csrB, gName)*reverseX ? hcsr($csrA, gName) : hcsr($csrB, gName)
		// if (isImage)																			// pin-pointing the minimum with the cursors can lead to unexpected behavior
			// s.yBase = zcsr($csrA, gName) < zcsr($csrB, gName) ? zcsr($csrA, gName) : zcsr($csrB, gName)
		// else
			// s.yBase = vcsr($csrA, gName) < vcsr($csrB, gName) ? vcsr($csrA, gName) : vcsr($csrB, gName)
		// endif
	endif
	// ++++++++++++++++++++++++++++++++++++ do the fit +++++++++++++++++++++++++++++++++++++++++++
	String printStr = "", execStr = "", cwStr = ""
	if (StringMatch(fitType, "FuncFit"))
		FUNCREF dummyFunc_prepareCoef coefFunc=$(fitFunc+CoefGen_SUFFIX)						// prepare coef wave
		int error = coefFunc(s)
		if (error || !WaveExists(s.cw))
			Print "Guess for "+fitFunc+" failed."
			return -1
		endif
		
		printStr = "•Make/D/O/N="+num2str(DimSize(s.cw,0))+" W_coef={"
		for (i=0; i<DimSize(s.cw,0); i++)
			printStr += num2str(s.cw[i]) + ","
		endfor
		printStr = RemoveEnding(printStr,",")+"}"
		Print printStr																			// print coef creation command
		cwStr = GetWavesDataFolder(s.cw,2)+" "
	else
		K0 = s.holdMode == 1 ? 0 : s.yBase
		Printf "•K0 = %g;\r", K0
	endif
	
	String errBarInfo = StringByKey("ERRORBARS",TraceInfo(gName, trace,0))						// find error bar wave
	if (strlen(errBarInfo))
		sscanf StringByKey("wave",errBarInfo,"=",","), "(%s,", execStr
		Wave/Z ebWave = $execStr
	endif
	
	String dataName = "", ebName = "", xwName = "", ywName = ""
	if (WaveExists(ebWave) && doWeight)															// add x-wave and error info
		sPrintf ebName, "/W=%s[%d,%d]",GetWavesDataFolder(ebWave,2), s.pntMin, s.pntMax
	endif
	if (WaveExists(s.xw))
		sPrintf xwName, "/X=%s[%d,%d]",GetWavesDataFolder(s.xw,2), s.pntMin, s.pntMax
	endif
	if (isImage)																				// the data to fit
		sPrintf dataName, "%s[%d,%d][%d,%d]",GetWavesDataFolder(s.data,2), s.pntMin, s.pntMax, s.pntMinY, s.pntMaxY
	else
		sPrintf dataName, "%s[%d,%d]%s",GetWavesDataFolder(s.data,2), s.pntMin, s.pntMax, SelectString(isTraceOf2D,"","["+num2str(s.pntCol)+"]")
	endif
	
	String doTxtbox	= SelectString(s.doTextbox,"","/TBOX="+num2str(textboxOpt))					// construct the fit command
	String dofullR	= SelectString(s.plotFull,"","/X=1")
	String doHold	= SelectString(s.holdMode>0,"","/H=\""+s.holdStr+"\"")
	sPrintf execStr, "•%s/M=2%s%s%s %s %s%s /D%s%s\r", fitType, doTxtbox, dofullR, doHold, fitFunc, cwStr, dataName, xwName, ebName
	Print execStr
	Execute/Z execStr																			// do the fit
	Wave/Z fit = $("fit_"+NameOfWave(s.data))
	if (V_flag)
		Print "Execution of the fit failed."
		if (WaveExists(fit))																	// invalidate any previous fit curve
			fit = NaN
		endif
		return -1
	endif

	if (StringMatch(fitType, "CurveFit"))
		Wave s.cw = W_coef
	endif
	
	if (StringMatch(fitFunc, "Line"))															// the line function does not output M_Covar
		Wave s.sw = W_sigma
	else
		Wave s.sw = M_Covar
	endif
	
	if (Exists(fitFunc+derivedVal_SUFFIX) == 6)
		FUNCREF dummyFunc_derivedVals outputFunc=$(fitFunc+derivedVal_SUFFIX)					// output additional parameters
		printStr = outputFunc(s)
		if (strlen(printStr))
			Print printStr
		endif
		if (WaveExists(fit))
			Note fit, printStr
		endif
	endif
	
	if (doFitAlign && WaveExists(fit))															// ver. 1.06 - added option for aligned fit traces
		String fitTrace = PossiblyQuoteName(NameOfWave(fit))
		if (FindListItem(fitTrace, TraceNameList(gName,";",1)) != -1)
			Variable xOff = 0, yOff = 0, xScale = 0, yScale = 0									// get scaling of source trace
			sscanf StringByKey("offset(x)", 	TraceInfo(gName,trace,0), "="), "{%f,%f}", xOff, yOff
			sscanf StringByKey("muloffset(x)",	TraceInfo(gName,trace,0), "="), "{%f,%f}", xScale, yScale
			ModifyGraph/Z/W=$gName offset($fitTrace)={xOff,yOff}, muloffset($fitTrace)={xScale,yScale}
		endif
	endif
	
	String remList = ""
	if (doCleanUp)																				// remove working variables and waves
		remList = "V_FitOptions;V_FitTol;V_tol;V_chisq;V_q;V_siga;V_sigb;V_Rab;V_Pr;V_r2;V_npnts;V_nterms;V_nheld;V_numNaNs;V_numINFs;"
		remList+= "V_FitError;V_FitQuitReason;V_FitIterStart;V_FitMaxIters;V_FitNumIters;"
		remList+= "V_startRow;V_startCol;V_startLayer;V_startChunk;V_endRow;V_endCol;V_endLayer;V_endChunk;"
		for(i=0; i<ItemsInList(remList); i++)
			KillVariables/Z $StringFromList(i,remList)
		endfor
		remList = "S_Info;"
		for(i=0; i<ItemsInList(remList); i++)
			KillStrings/Z $StringFromList(i,remList)
		endfor
		KillWaves/Z W_coef, W_sigma, M_Covar, W_ParamConfidenceInterval
	endif
	return 0
End

//________________________________________________________________________________________________
//									Default Fit-Function Output
//________________________________________________________________________________________________

Function/S Line_derivedVals(STRUCT SuperQuickFitStruct &s)
	NVAR V_rab
	Variable xCross = -s.cw[0]/s.cw[1]
	Variable xCross_err	= xCross * sqrt( (s.sw[0]/s.cw[0])^2 + (s.sw[1]/s.cw[1])^2 + 2*(V_rab*s.sw[0]*s.sw[1])/(s.cw[0]*s.cw[1]) )
	xCross_err = numtype(xCross_err) != 0 ? 0 : xCross_err
	String printStr = ""
	sPrintf printStr, "Derived values:\r\tX-intercept\t= %g ± %g", xCross, xCross_err
	if (s.doTextbox)
		AppendText "\t"+printStr
	endif
	return printStr
End

Function/S Gauss_derivedVals(STRUCT SuperQuickFitStruct &s)
	Variable pArea	= s.cw[1]*abs(s.cw[3])*sqrt(Pi)
	Variable FWHM	= abs(s.cw[3])*2*sqrt(ln(2))
	Variable pArea_err	= pArea * sqrt( (s.sw[3][3]/s.cw[3]^2) + (s.sw[1][1]/s.cw[1]^2) + 2*s.sw[1][3]/(s.cw[1]*s.cw[3]) )
	Variable FWHM_err	= sqrt(s.sw[3][3])*2*sqrt(ln(2))
	String printStr = ""
	sPrintf printStr, "Derived values:\r\tArea\t= %g ± %g\r\tFWHM\t= %g ± %g", pArea, pArea_err, FWHM, FWHM_err
	if (s.doTextbox)
		AppendText "\t"+printStr
	endif
	return printStr
End

Function/S Lor_derivedVals(STRUCT SuperQuickFitStruct &s)
	Variable Height	= s.cw[1]/abs(s.cw[3])
	Variable FWHM	= 2*sqrt(abs(s.cw[3]))
	Variable pArea	= pi*s.cw[1]/sqrt(abs(s.cw[3]))
	Variable Height_err	= Height * sqrt( (s.sw[3][3]/s.cw[3]^2) + (s.sw[1][1]/s.cw[1]^2) + 2*s.sw[1][3]/(s.cw[1]*s.cw[3]) )
	Variable FWHM_err	= FWHM/s.cw[3] * sqrt(s.sw[3][3])
	Variable pArea_err	= sqrt( s.sw[1][1] * pi^2/abs(s.cw[3]) + s.sw[3][3] * (pi^2*s.cw[1]^2)/(4*abs(s.cw[3])^3) - 2*abs(s.sw[1][3])*(pi*s.cw[1])/(2*abs(s.cw[3])^2) )
	String printStr = ""
	sPrintf printStr, "Derived values:\r\tHeight\t= %g ± %g\r\tArea\t= %g ± %g\r\tFWHM\t= %g ± %g", Height, Height_err, pArea, pArea_err, FWHM, FWHM_err
	if (s.doTextbox)
		AppendText "\t"+printStr
	endif
	return printStr
End

Function/S Voigt_derivedVals(STRUCT SuperQuickFitStruct &s)
	Variable FWHM_err = 0, Height_err = 0, C1 = 0.5346, C2 = 0.2166, k = 2*sqrt(ln(2)/pi)
	Variable shape	= s.cw[4]*sqrt(ln(2)), lorW = s.cw[4]*s.cw[3]
	
	Variable height	= k*s.cw[1]/abs(s.cw[3]) * exp(shape^2)*erfc(shape)
	Variable FWHM = (C1*s.cw[4] + sqrt(C2*s.cw[4]^2 + 1))*abs(s.cw[3])

	Height_err +=   s.sw[1][1] * (  k/abs(s.cw[3])*exp(shape^2)*erfc(shape) )^2
	Height_err +=   s.sw[2][2] * ( (sqrt(pi)*shape*exp(shape^2)*erfc(shape) - 1)*k^2*s.cw[1]/abs(s.cw[3]) )^2
	Height_err +=   s.sw[3][3] * (  k*s.cw[1]/abs(s.cw[3])^2*exp(shape^2)*erfc(shape) )^2
	Height_err += 2*s.sw[1][3] *  k^2*s.cw[1]/abs(s.cw[3])^3*(exp(shape^2)*erfc(shape))^2
	Height_err -= 2*s.sw[1][2] *  (sqrt(pi)*shape*exp(shape^2)*erfc(shape) - 1)*k^3*s.cw[1]  /abs(s.cw[3])   *exp(shape^2)*erfc(shape)
	Height_err -= 2*s.sw[2][3] *  (sqrt(pi)*shape*exp(shape^2)*erfc(shape) - 1)*k^3*s.cw[1]^2/abs(s.cw[3])^2 *exp(shape^2)*erfc(shape)
	Height_err = sqrt(Height_err)
	
	FWHM_err += s.sw[3][3] * (  C1*s.cw[4] + sqrt(C2*s.cw[4]^2 + 1) )^2
	FWHM_err += s.sw[4][4] * ( (C1*s.cw[4] + C2*s.cw[4]/sqrt(C2*s.cw[4]^2 + 1))*abs(s.cw[3]) )^2
	FWHM_err += 2*abs(s.sw[3][4])*(C1*s.cw[4] + sqrt(C2*s.cw[4]^2 + 1))*(C1*s.cw[4] + C2*s.cw[4]/sqrt(C2*s.cw[4]^2 + 1))*abs(s.cw[3])
	FWHM_err = sqrt(FWHM_err)
	
	String printStr = ""
	sPrintf printStr, "Derived values:\r\tHeight\t= %g ± %g\r\tFWHM\t= %g ± %g", Height, Height_err, FWHM, FWHM_err
	if (s.doTextbox)
		AppendText "\t"+printStr
	endif
	return printStr
End

//________________________________________________________________________________________________
//									Custom Curve Fit Functions
//________________________________________________________________________________________________

Function FermiEdge_Line(Wave w, Variable E) : FitFunc
	//CurveFitDialog/ Equation:
	//CurveFitDialog/ f(E) = y0 + (m*E+A) / (exp((E-Ef)/kT)+1)
	//CurveFitDialog/ 
	//CurveFitDialog/ End of Equation
	//CurveFitDialog/ Independent Variables 1
	//CurveFitDialog/ E
	//CurveFitDialog/ Coefficients 5
	//CurveFitDialog/ w[0] = y0
	//CurveFitDialog/ w[1] = Ef
	//CurveFitDialog/ w[2] = kT
	//CurveFitDialog/ w[3] = A
	//CurveFitDialog/ w[4] = m
	return w[0] + (w[4]*E+w[3]) / (exp((E-w[1])/w[2])+1)
End

Function FermiEdge_Line_prepareCoef(STRUCT SuperQuickFitStruct &s)
	Duplicate/free/RMD=[][s.pntCol] s.data, dif													// support for 2D trace data
	Differentiate dif/D=dif
	Variable smthFact = round(abs(s.pntMax-s.pntMin)/25)
	smthFact = limit(smthFact, 1, 2^15-1)
	Smooth smthFact, dif
	WaveStats/Q/R=[s.pntMin,s.pntMax] dif
	Variable Ef
	if (WaveExists(s.xw))
		Ef = s.xw[V_minRowLoc]
	else
		Ef = V_minloc
	endif
	Make/D/O/N=2 W_coef = {(s.holdMode == 1 ? 0 : s.yBase),Ef,0.025,WaveMax(s.data),-1}
	Wave s.cw = W_coef
	return 0
End

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

Function LineX(Wave w, Variable x) : FitFunc
	//CurveFitDialog/ Equation:
	//CurveFitDialog/ f(x) = m*(x-x0)
	//CurveFitDialog/ End of Equation
	//CurveFitDialog/ Independent Variables 1
	//CurveFitDialog/ x
	//CurveFitDialog/ Coefficients 2
	//CurveFitDialog/ w[0] = x0
	//CurveFitDialog/ w[1] = m
	return w[1]*(x-w[0])
End

Function LineX_prepareCoef(STRUCT SuperQuickFitStruct &s)
	Duplicate/free/RMD=[][s.pntCol] s.data, dif													// support for 2D trace data
	Differentiate dif/D=dif
	WaveStats/Q/R=[s.pntMin,s.pntMax] dif
	Variable slope = (WaveMax(s.data) - WaveMin(s.data)) > 0 ? V_max : V_min
	Make/D/O/N=2 W_coef = {(s.holdMode == 1 ? 0 : s.xBase),slope}
	Wave s.cw = W_coef
	return 0
End

Function/S LineX_derivedVals(STRUCT SuperQuickFitStruct &s)
	Variable yCross = -s.cw[1]*s.cw[0]
	Variable yCross_err	= yCross * sqrt( (s.sw[0][0]/s.cw[0]^2) + (s.sw[1][1]/s.cw[1]^2) + 2*s.sw[0][1]/(s.cw[0]*s.cw[1]) )
	String printStr = ""
	sPrintf printStr, "Derived values:\r\tY-intercept\t= %g ± %g", yCross, (numtype(yCross_err) != 0 ? 0 : yCross_err)
	if (s.doTextbox)
		AppendText "\t"+printStr
	endif
	return printStr
End

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

Function PAD_Beta(Wave w, Variable x) : FitFunc
	//CurveFitDialog/ Equation:
	//CurveFitDialog/ f(theta) = scale*(1+beta*legendreA(2,0,cos((theta-theta_0)*(Pi/180))))
	//CurveFitDialog/ End of Equation
	//CurveFitDialog/ Independent Variables 1
	//CurveFitDialog/ theta
	//CurveFitDialog/ Coefficients 3
	//CurveFitDialog/ w[0] = scale
	//CurveFitDialog/ w[1] = beta
	//CurveFitDialog/ w[2] = theta_0
	return w[0]*(1+w[1]*legendreA(2,0,cos((x-w[2])*(Pi/180))))
End

Function PAD_Beta_prepareCoef(STRUCT SuperQuickFitStruct &s)
	s.holdStr = "001"
	Make/D/O/N=3 W_coef = {WaveMax(s.data)/3,2,0}
	Wave s.cw = W_coef
	return 0
End

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

Function DoubleGauss(Wave w, Variable x) : FitFunc
	//CurveFitDialog/ Equation:
	//CurveFitDialog/ f(x) = y0 + Area1 * Gauss(x,Loc1,Width1) + Area2 * Gauss(x,Loc2,Width2)
	//CurveFitDialog/ End of Equation
	//CurveFitDialog/ Independent Variables 1
	//CurveFitDialog/ x
	//CurveFitDialog/ Coefficients 7
	//CurveFitDialog/ w[0] = y0
	//CurveFitDialog/ w[1] = Area1
	//CurveFitDialog/ w[2] = Area2
	//CurveFitDialog/ w[3] = Loc1
	//CurveFitDialog/ w[4] = Loc2
	//CurveFitDialog/ w[5] = Width1
	//CurveFitDialog/ w[6] = Width2
	return w[0]+ w[1] * Gauss(x,w[3],w[5]) + w[2] * Gauss(x,w[4],w[6])
End

Function DoubleGauss_prepareCoef(STRUCT SuperQuickFitStruct &s)
	Duplicate/free/RMD=[][s.pntCol] s.data, tempData											// support for 2D trace data
	Variable minVal = s.holdMode == 1 ? 0 : s.yBase
	FindPeak/P/B=5/Q/R=[s.pntMin,s.pntMax]/M=(wavemax(tempData)/10+minVal) tempData
	Variable pLoc1 = V_peakloc, pWidth1 = V_PeakWidth, pHeight1 = V_PeakVal
	if (V_flag)
		return -1
	endif
		FindPeak/P/B=5/Q/R=[s.pntMin,pLoc1-5]/M=(wavemax(tempData)/10+minVal) tempData			// look left of peak
	if (V_flag)
		FindPeak/P/B=5/Q/R=[pLoc1+5,s.pntMax]/M=(wavemax(tempData)/10+minVal) tempData			// look right of peak
	endif
	Variable pLoc2 = V_peakloc, pWidth2 = V_PeakWidth, pHeight2 = V_PeakVal
	if (V_flag)
		return -1
	endif
	if (WaveExists(s.xw))
		pLoc1 = s.xw[pLoc1]
		pLoc2 = s.xw[pLoc2]
		pWidth1 *= abs((s.xw[numpnts(s.xw)-1]-s.xw[0])/(numpnts(s.xw)-1))
		pWidth2 *= abs((s.xw[numpnts(s.xw)-1]-s.xw[0])/(numpnts(s.xw)-1))
	else
		pLoc1 = pnt2x(s.data,pLoc1)
		pLoc2 = pnt2x(s.data,pLoc2)
		pWidth1 *= abs(DimDelta(s.data,0))
		pWidth2 *= abs(DimDelta(s.data,0))
	endif
	Make/D/O/N=7 W_coef = {minVal,pHeight1,pHeight2,pLoc1,pLoc2,pWidth1,pWidth2}
	Wave s.cw = W_coef
	return 0
End

Function/S DoubleGauss_derivedVals(STRUCT SuperQuickFitStruct &s)
	Variable FWHM1		= abs(s.cw[5])*2*sqrt(2*ln(2)),			FWHM2		= abs(s.cw[6])*2*sqrt(2*ln(2))
	Variable FWHM1_err	= sqrt(s.sw[5][5])*2*sqrt(2*ln(2)),		FWHM2_err	= sqrt(s.sw[6][6])*2*sqrt(2*ln(2))
	Variable Amp1		= s.cw[1]/(abs(s.cw[5])*sqrt(2*Pi)),	Amp2		= s.cw[2]/(abs(s.cw[6])*sqrt(2*Pi))
	Variable Amp1_err	= Amp1 * sqrt( (s.sw[5][5]/s.cw[5]^2) + (s.sw[1][1]/s.cw[1]^2) + 2*s.sw[1][5]/(s.cw[1]*s.cw[5]) )
	Variable Amp2_err	= Amp2 * sqrt( (s.sw[6][6]/s.cw[6]^2) + (s.sw[1][1]/s.cw[1]^2) + 2*s.sw[1][6]/(s.cw[1]*s.cw[6]) )
	String printStr = ""
	sPrintf printStr, "Derived values:\r\tAmp1\t= %g ± %g\r\tAmp2\t= %g ± %g\r\tFWHM1\t= %g ± %g\r\tFWHM2\t= %g ± %g", Amp1, Amp1_err, Amp2, Amp2_err, FWHM1, FWHM1_err, FWHM2, FWHM2_err
	if (s.doTextbox)
		AppendText "\t"+printStr
	endif
	return printStr
End

// ####################################### Helper functions ######################################

static Function getShapeFromData(STRUCT SuperQuickFitStruct &s,Variable &loc, Variable &amp, Variable &base, Variable &lwidth, Variable &rwidth)	// finds basic parameters like hieght, width and location from input data
	Duplicate/free/RMD=[][s.pntCol] s.data, tempData											// support for 2D trace data
	WaveStats/Q/R=[s.pntMin,s.pntMax] tempData
	amp  = abs(V_max) > abs(V_min) ? V_max-V_min : V_min-V_max
	base = s.holdMode == 1 ? 0 : s.yBase
	if (WaveExists(s.xw))
		loc = amp > 0 ? s.xw[V_maxRowloc] : s.xw[V_minRowloc]
	else
		loc = amp > 0 ? V_maxloc : V_minloc
	endif
	lwidth = NaN; rwidth = NaN
	// +++++++++++++++++++++++++++++++++ width and asymmetry +++++++++++++++++++++++++++++++++++++
	Make/FREE/D/N=2 Levels
	Variable reverseX, avg = ceil((s.pntMax-s.pntMin)*0.05)
	if(WaveExists(s.xw))
		reverseX = sign(s.xw[numpnts(s.xw)-1]-s.xw[0])
		FindLevels/Q/R=[s.pntMin,s.pntMax]/D=Levels/B=(limit(avg,1,20))/P tempData, amp/2+base	// find the half-maximum positions
		if (V_LevelsFound == 2)
			if (Levels[0] > 0 && Levels[1] > 0 && Levels[0] < numpnts(s.xw) && Levels[1] < numpnts(s.xw))
				lwidth = (loc - s.xw[Levels[0]])*reverseX
				rwidth = (s.xw[Levels[1]] - loc)*reverseX
			endif
		endif
	else
		reverseX = sign(DimDelta(s.data,0))
		FindLevels/Q/R=[s.pntMin,s.pntMax]/D=Levels/B=(limit(avg,1,20))   tempData, amp/2+base
		if (V_LevelsFound == 2)
			lwidth = (loc - Levels[0])*reverseX
			rwidth = (Levels[1] - loc)*reverseX
		endif
	endif
	return 0
End

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static Function getPeakParameters(String PeakFunc, Wave cw, Variable &amp, Variable &loc, Variable &fwhm)	// outputs peak height, location and FWHM from a set of initial guesses
	Variable isPos = amp > 0, start_w = fwhm
	Optimize/Q/A=(isPos)/L=(loc-5*start_w)/H=(loc+5*start_w) $PeakFunc,cw						// find peak maximum numerically
	amp  = isPos ? V_max : V_min
	loc  = isPos ? V_maxLoc : V_minLoc
	fwhm = AnalysisTools#getFWHM(PeakFunc,cw, amp, loc, start_w)
	return 0
End

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static Function getFWHM(String PeakFunc, Wave cw, Variable amp, Variable loc, Variable start_w)	// finds the approximate FWHM numerically
	Variable highRange, lowRange, highHW, LowHW
	highRange = amp > 0 ? loc : loc+5*start_w
	lowRange  = amp > 0 ? loc+5*start_w : loc
	FindRoots/Q/H=(highRange)/L=(lowRange)/T=1e-15/Z=(amp/2) $PeakFunc,cw
	highHW = V_flag ? NaN : V_Root
	highRange = amp > 0 ? loc : loc-5*start_w
	lowRange  = amp > 0 ? loc-5*start_w : loc
	FindRoots/Q/H=(highRange)/L=(lowRange)/T=1e-15/Z=(amp/2) $PeakFunc,cw
	LowHW = V_flag ? NaN : V_Root
	return abs(highHW - LowHW)
End

// ####################################### EMG Peak function #####################################

Function ExpModGauss(Wave cw, Wave yw, Wave xw)													// the fit function
	Make/FREE/D/N=4 peak_cw = cw[p]
	Variable dummy = MPFXEMGPeak(peak_cw, yw, xw)
	yw += cw[4]
End
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Function ExpModGauss_prepareCoef(STRUCT SuperQuickFitStruct &s)									// the coef generator
	Make/D/O/N=5 W_coef	= NaN		// [0]: x0, [1]: width, [2]: amplitude, [3]: asymmetry, [4]: y0
	Variable, loc, amp, base, lwidth, rwidth
	AnalysisTools#getShapeFromData(s, loc, amp, base, lwidth, rwidth)							// gets location, amplitude, baseline and width from the data
	
	W_coef[0] = loc
	W_coef[2] = amp
	W_coef[4] = base
	if (numtype(lwidth+rwidth) == 0)															// half-maximum positions found
		W_coef[0] += (lwidth - rwidth)/2														// move the peak over a bit
		if (lwidth > rwidth)																	// left tail is longer
			W_coef[1] = rwidth
			W_coef[3] = -3*abs(lwidth - rwidth)													// updated ExpTau calculation to work well with asymmetric peak edit
			W_coef[3] = min(W_coef[3],-0.2*W_coef[1])											// ExpModGauss needs some minimal asymmetry
		else
			W_coef[1] = lwidth
			W_coef[3] = 3*abs(lwidth - rwidth)
			W_coef[3] = max(W_coef[3],0.2*W_coef[1])
		endif
		Variable widthRatio = abs(W_coef[3])/W_coef[1]
		W_coef[2] = W_coef[2]/(1-(widthRatio/(widthRatio+1.2))^2)								// totally empirically determined
	else
		W_coef[1] = 1																			// may not be good guesses
		W_coef[3] = 0.2
	endif
	
	Wave s.cw = W_coef
	s.holdStr = "00001"																			// may hold baseline
	return 0
End
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Function/S ExpModGauss_derivedVals(STRUCT SuperQuickFitStruct &s)								// calculate derived parameters
	Variable amp = s.cw[2], loc = s.cw[0], fwhm = abs(s.cw[1])									// must provide (rough) guesses
	AnalysisTools#getPeakParameters("AnalysisTools#EMGFunc", s.cw, amp, loc, fwhm)
	
	Variable pArea = s.cw[1]*s.cw[2]*sqrt(2*pi)													// the peak area can be calculated analytically
	Variable pArea_err = sqrt(s.sw[1][1]*(s.cw[2]*sqrt(2*pi))^2 + s.sw[2][2]*(s.cw[1]*sqrt(2*pi))^2 + 2*abs(s.sw[1][2])*(s.cw[1]*s.cw[2]*2*pi))
	
	String printStr = ""
	sPrintf printStr, "Area\t= %.5g ± %.5g\r\tHeight\t= %.7g\r\tLocation\t= %.7g\r\tFWHM\t= %.7g", pArea, pArea_err, amp, loc, fwhm
	if (s.doTextbox)
		AppendText "\t"+printStr																// also add derived parameters into text-box
	endif
	Print "K0 = x0, K1 = width, K2 = amplitude, K3 = asymmetry, K4 = y0"						// print explanation
	return "Derived values:\r\t"+printStr
End
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static Function EMGFunc(Wave w, Variable x)														// function for finding peak values
	Make/D/FREE/N=1 ytemp, xtemp=x
	Variable dummy = MPFXEMGPeak(w, ytemp, xtemp)
	return ytemp[0]
End

// ####################################### ExpConvExp function #####################################

Function ExpConvExp(Wave cw, Wave yw, Wave xw)													// the fit function
	Make/FREE/D/N=4 peak_cw = cw[p]
	Variable dummy = MPFXExpConvExpPeak(peak_cw, yw, xw)
	yw += cw[4]
End
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Function ExpConvExp_prepareCoef(STRUCT SuperQuickFitStruct &s)									// the coef generator
	Make/D/O/N=5 W_coef	= NaN	// [0]: x0, [1]: height, [2]: 1st inverse decay constant, [3]: 2nd inverse decay constant, [4]: y0
	Variable, loc, amp, base, lwidth, rwidth
	AnalysisTools#getShapeFromData(s, loc, amp, base, lwidth, rwidth)
	
	W_coef[0] = loc
	W_coef[1] = amp
	W_coef[4] = base
	if (numtype(lwidth+rwidth) == 0)															// half-maximum positions found
		W_coef[2] = 1/lwidth
		W_coef[3] = 1/rwidth
	else
		W_coef[2] = 1/(0.04*(s.pntMax-s.pntMin)*abs(DimDelta(s.data,0)))						// may not be good guesses
		W_coef[3] = 1/(0.06*(s.pntMax-s.pntMin)*abs(DimDelta(s.data,0)))
	endif
	if (W_coef[2] >= W_coef[3])																	// coefficients should not be exactly the same
		W_coef[3] = 1.2*W_coef[2]																// make sure that decay coefficients are in the right order
	endif
	W_coef[1] *= 1.6																			// make the peak a bit higher
	W_coef[0] -= (ln(W_coef[2]/W_coef[3])/(W_coef[2]-W_coef[3]))								// shift the peak over
	
	Wave s.cw = W_coef
	s.holdStr = "00001"																			// may hold baseline
	return 0
End
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Function/S ExpConvExp_derivedVals(STRUCT SuperQuickFitStruct &s)								// calculate derived parameters
	Variable amp, loc, fwhm
	AnalysisTools#getExpConvExpVals(s.cw, amp, loc, fwhm)

	Variable pArea = s.cw[1]/s.cw[3]															// the peak area can be calculated analytically
	Variable pArea_err = sqrt(s.sw[1][1]/(s.cw[3])^2 + s.sw[3][3]*(s.cw[1]/s.cw[3]^2)^2 + 2*abs(s.sw[1][3])*(s.cw[1]/s.cw[3]^3))

	Variable K3 = s.cw[3], K3_err = sqrt(s.sw[3][3])
	Variable K2_mod = s.cw[2]-s.cw[3]															// add alternative coefficients
	Variable K2_err = sqrt(s.sw[2][2] + s.sw[3][3] + 2*abs(s.sw[2][3]))
	Variable K1_mod = (s.cw[2]*s.cw[1])/K2_mod
	
	Variable d1 = s.cw[2]/K2_mod, d2 = (s.cw[3]*s.cw[1])/K2_mod^2, d3 = (s.cw[2]*s.cw[1])/K2_mod^2				// derivatives
	Variable K1_err = s.sw[1][1]*(d1)^2 + s.sw[2][2]*(d2)^2 + s.sw[3][3]*(d3)^2
	K1_err += 2*abs(s.sw[1][3])*d1*d3 + 2*abs(s.sw[1][2])*d1*d2 + 2*abs(s.sw[2][3])*d2*d3						// side terms
	K1_err = sqrt(K1_err)

	String baseStr = "", advStr = ""
	sPrintf baseStr, "Area\t= %.5g ± %.5g\r\tHeight\t= %.7g\r\tLocation\t= %.7g\r\tFWHM\t= %.7g", pArea, pArea_err, amp, loc, fwhm
	sPrintf advStr,  "Virt. amplitude\t= %.5g ± %.5g\r\tInv. of rise\t= %.5g ± %.5g\r\tInv. of decay\t= %.5g ± %.5g", K1_mod, K1_err, K2_mod, K2_err, K3, K3_err
	if (s.doTextbox)
		AppendText "\t"+baseStr + "\r\t-----\r\t" + advStr														// also add derived parameters into text-box
	endif
	Print "K0 = location, K1 = amplitude,  K2 = 1st inv. decay const., K3 = 2nd inv. decay const., K4 = baseline"	// print explanation
	return "Derived values:\r\t"+baseStr + "\r\t-----\r\t" + advStr	
End
// ++++++++++++++++++++++++++++++ recreated from Multipeak Fit +++++++++++++++++++++++++++++++++++
static Function getExpConvExpVals(Wave cw, Variable &amp, Variable &loc, Variable &fwhm)
	if (cw[2] == cw[3])
		amp = cw[1]/e
		loc = 1/cw[3] + cw[0]
	else
		amp = (cw[1]*cw[2] * ((cw[2]/cw[3])^(-(cw[2]/(cw[2]-cw[3]))) - (cw[2]/cw[3])^(-(cw[3]/(cw[2] - cw[3]))))) / (-cw[2]+cw[3])
		loc = ln(cw[2]/cw[3])/(cw[2]-cw[3]) + cw[0]
	endif
	fwhm  = AnalysisTools#getFWHM("AnalysisTools#ExpConvExpFunc",cw, amp, loc, abs(1/cw[2]+1/cw[3]))
	return 0
End
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static Function ExpConvExpFunc(Wave w, Variable x)												// function for finding peak values
	Make/D/FREE/N=1 ytemp, xtemp=x
	Variable dummy = MPFXExpConvExpPeak(w, ytemp, xtemp)
	return ytemp[0]
End